/*
Copyright (c) Members of the EGEE Collaboration. 2004.
See http://www.eu-egee.org/partners/ for details on the
copyright holders.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND,
either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/


// $Id: ism_utils.cpp,v 1.1.2.7.2.1.2.2 2010/04/07 14:02:46 mcecchi Exp $

#include <vector>

#include <boost/bind.hpp>
#include "ism_utils.h"
#include "signal_handling.h"
#include "dynamic_library.h"
#include "glite/wms/common/configuration/Configuration.h"
#include "glite/wms/common/configuration/WMConfiguration.h"
#include "glite/wms/common/configuration/NSConfiguration.h"
#include "glite/wms/common/configuration/CommonConfiguration.h"
#include "glite/wms/common/logger/logger_utils.h"
#include "glite/wmsutils/classads/classad_utils.h"
#include "glite/wms/ism/ism.h"

#include "glite/wms/ism/purchaser/ism-purchaser.h"
#include "glite/wms/ism/purchaser/ism-ii-purchaser.h"
//#include "glite/wms/ism/purchaser/ism-cemon-purchaser.h"
//#include "glite/wms/ism/purchaser/ism-cemon-asynch-purchaser.h"
#include "glite/wms/ism/purchaser/ism-file-purchaser.h"
//#include "glite/wms/ism/purchaser/ism-rgma-purchaser.h"

namespace configuration = glite::wms::common::configuration;
namespace ca = glite::wmsutils::classads;
namespace purchaser = glite::wms::ism::purchaser;
namespace ism = glite::wms::ism;

namespace glite {
namespace wms {
namespace manager {
namespace main {

namespace {

typedef boost::shared_ptr<server::DynamicLibrary> DynamicLibraryPtr;

DynamicLibraryPtr make_dll(std::string const& lib_name)
{
  return DynamicLibraryPtr(new server::DynamicLibrary(lib_name));
}

void load_dlls_and_create_purchasers(
  std::vector<DynamicLibraryPtr>& dlls,
  std::vector<
    boost::shared_ptr<purchaser::ism_purchaser>
  >& purchasers
)
{
  configuration::Configuration const& config(
    *configuration::Configuration::instance()
  );

  purchaser::exit_predicate_type const exit_predicate(
    server::received_quit_signal
  );
  purchaser::skip_predicate_type const skip_predicate(
    purchaser::is_in_black_list(config.wm()->ism_black_list())
  );

  ism::purchaser::exec_mode_t mode;
  if (config.wm()->ism_threads()) {
    mode = purchaser::loop;
  } else {
    mode = purchaser::once;
  }

  // bool const purchasing_from_rgma_is_enabled(
  //   config.wm()->enable_purchasing_from_rgma()
  // );
  // std::vector<std::string> const cemon_services(
  //   config.wm()->ce_monitor_services()
  // );
  // bool const purchasing_from_cemon_is_enabled(!cemon_services.empty());
  // bool const purchasing_from_cemon_asynch_is_enabled(
  //   config.wm()->ce_monitor_asynch_port() > 0
  // );

  purchaser::ii::create_entry_update_fn_t* ii_update_function = 0;
  //purchaser::cemon::create_entry_update_fn_t* cemon_update_function = 0;
  //purchaser::rgma::create_entry_update_fn_t* rgma_update_function = 0;

  {

    Info("loading II purchaser");

    std::string const dll_name("libglite_wms_ism_ii_purchaser.so");
    DynamicLibraryPtr dll(make_dll(dll_name));

    std::string const create_function_name("create_ii_purchaser");
    purchaser::ii::create_t* create_function;
    dll->lookup(create_function_name, create_function);

    std::string const destroy_function_name("destroy_ii_purchaser");
    purchaser::ii::destroy_t* destroy_function;
    dll->lookup(destroy_function_name, destroy_function);

    Info("first II purchasing in progress...");
    boost::shared_ptr<purchaser::ism_purchaser> first_ii_purchasing(
      create_function(
        config.ns()->ii_contact(),
        config.ns()->ii_port(),
        config.ns()->ii_dn(),
        config.ns()->ii_timeout(),
        config.wm()->ism_ii_ldapcefilter_ext(),
        config.wm()->ism_ii_ldapsearch_async(),
        purchaser::once, // always once here
        0,
        exit_predicate,
        skip_predicate
      ),
      destroy_function
    );
    (*first_ii_purchasing)(); // first time is done synchronously

    boost::shared_ptr<purchaser::ism_purchaser> purchaser(
      create_function(
        config.ns()->ii_contact(),
        config.ns()->ii_port(),
        config.ns()->ii_dn(),
        config.ns()->ii_timeout(),
        config.wm()->ism_ii_ldapcefilter_ext(),
        config.wm()->ism_ii_ldapsearch_async(),
        mode,
        config.wm()->ism_ii_purchasing_rate(),
        exit_predicate,
        skip_predicate
      ),
      destroy_function
    );

    dlls.push_back(dll);
    purchasers.push_back(purchaser);

    std::string const update_function_name("create_ii_entry_update_fn");
    dll->lookup(update_function_name, ii_update_function);
  }

  /****************************************************************************
  if (purchasing_from_rgma_is_enabled) {

    Info("loading rgma purchaser");

    std::string const dll_name("libglite_wms_ism_rgma_purchaser.so");
    DynamicLibraryPtr dll(make_dll(dll_name));

    std::string const create_function_name("create_rgma_purchaser");
    purchaser::rgma::create_t* create_function;
    dll->lookup(create_function_name, create_function);

    std::string const destroy_function_name("destroy_rgma_purchaser");
    purchaser::rgma::destroy_t* destroy_function;
    dll->lookup(destroy_function_name, destroy_function);

    boost::shared_ptr<purchaser::ism_purchaser> purchaser(
      create_function(
        config.wm()->rgma_query_timeout(),
        mode,
        config.wm()->rgma_consumer_ttl(),
        config.wm()->rgma_consumer_life_cycle(),
        config.wm()->ism_rgma_purchasing_rate(),
        exit_predicate,
        skip_predicate
      ),
      destroy_function
    );

    dlls.push_back(dll);
    purchasers.push_back(purchaser);

    std::string const update_function_name("create_rgma_entry_update_fn");
    dll->lookup(update_function_name, rgma_update_function);
  }

  if (purchasing_from_cemon_is_enabled) {

    Info("loading cemon purchaser");

    std::string const dll_name("libglite_wms_ism_cemon_purchaser.so");
    DynamicLibraryPtr dll(make_dll(dll_name));

    std::string const create_function_name("create_cemon_purchaser");
    purchaser::cemon::create_t* create_function;
    dll->lookup(create_function_name, create_function);

    std::string const destroy_function_name("destroy_cemon_purchaser");
    purchaser::cemon::destroy_t* destroy_function;
    dll->lookup(destroy_function_name, destroy_function);

    char const* certificate_path = ::getenv("GLITE_CERT_DIR");
    if (!certificate_path) {
      certificate_path = "/etc/grid-security/certificates";
    }

    boost::shared_ptr<purchaser::ism_purchaser> purchaser(
      create_function(
        config.common()->host_proxy_file(),
        certificate_path,
        cemon_services,
        "CE_MONITOR",
        120,
        mode,
        config.wm()->ism_cemon_purchasing_rate(),
        exit_predicate,
        skip_predicate
      ),
      destroy_function
    );

    dlls.push_back(dll);
    purchasers.push_back(purchaser);

    std::string const update_function_name("create_cemon_entry_update_fn");
    dll->lookup(update_function_name, cemon_update_function);
  }

  if (purchasing_from_cemon_asynch_is_enabled) {

    Info("loading asynch cemon purchaser");

    std::string const dll_name("libglite_wms_ism_cemon_asynch_purchaser.so");
    DynamicLibraryPtr dll(make_dll(dll_name));

    std::string const create_function_name("create_cemon_asynch_purchaser");
    purchaser::cemon_asynch::create_t* create_function;
    dll->lookup(create_function_name, create_function);

    std::string const destroy_function_name("destroy_cemon_asynch_purchaser");
    purchaser::cemon_asynch::destroy_t* destroy_function;
    dll->lookup(destroy_function_name, destroy_function);

    boost::shared_ptr<purchaser::ism_purchaser> purchaser(
      create_function(
        "CE_MONITOR",
        config.wm()->ce_monitor_asynch_port(),
        mode,
        config.wm()->ism_cemon_asynch_purchasing_rate(),
        exit_predicate,
        skip_predicate
      ),
      destroy_function
    );

    dlls.push_back(dll);
    purchasers.push_back(purchaser);
  }
  ****************************************************************************/
}

}

struct ISM_manager::Impl : boost::noncopyable
{
  ism::ism_type m_ism_1[2];
  ism::ism_type m_ism_2[2];
  ism::ism_mutex_type m_ism_mutex[2];
  std::vector<DynamicLibraryPtr> m_dlls;
  std::vector<boost::shared_ptr<purchaser::ism_purchaser> > m_purchasers;
};

ISM_manager::ISM_manager()
  : m_impl(new Impl)
{
  ism::set_ism(m_impl->m_ism_1, m_impl->m_ism_2, m_impl->m_ism_mutex, ism::ce);
  ism::set_ism(m_impl->m_ism_1, m_impl->m_ism_2, m_impl->m_ism_mutex, ism::se);

  load_dlls_and_create_purchasers(m_impl->m_dlls, m_impl->m_purchasers);
}

std::vector<
  boost::shared_ptr<purchaser::ism_purchaser>
>& ISM_manager::purchasers()
{
  return m_impl->m_purchasers;
}

}}}} // glite::wms::manager::main
